// Lighto's copy slope fix, edit by haya
// Feel free to steal and mangle

if copyslopefix_initialized then return end // Check if it's already been loaded.

local runscript = false

local FF_EXISTS = FF_EXISTS
local FF_SOLID = FF_SOLID
local FF_QUICKSAND = FF_QUICKSAND
local FF_SWIMMABLE = FF_SWIMMABLE
local MF2_OBJECTFLIP  = MF2_OBJECTFLIP
local FRACUNIT = FRACUNIT

local function L_GetSectorCeilingZAt(sector, x, y)
	if sector.c_slope then
		return P_GetZAt(sector.c_slope, x, y)
	end
	return sector.ceilingheight
end

local function L_GetFFloorTopZAt(ffloor, x, y)
	if ffloor.t_slope then
		return P_GetZAt(ffloor.t_slope, x, y)
	end
	return ffloor.topheight
end

local function L_GetFFloorBottomZAt(ffloor, x, y)
	if ffloor.b_slope then
		return P_GetZAt(ffloor.b_slope, x, y)
	end
	return ffloor.bottomheight
end

local function L_CeilingzAtPos(x, y, z, height)
	local sec = R_PointInSubsector(x, y).sector
	local ceilingz = L_GetSectorCeilingZAt(sec, x, y)
	for rover in sec.ffloors() do
		if not ((rover.flags & FF_EXISTS) and (rover.flags & (FF_SOLID|FF_QUICKSAND)))
			or (rover.flags & FF_SWIMMABLE) then continue end
		local thingtop = z + height
		
		local topheight    = L_GetFFloorTopZAt   (rover, x, y)
		local bottomheight = L_GetFFloorBottomZAt(rover, x, y)

		if rover.flags & FF_QUICKSAND then
			if thingtop > bottomheight and topheight > z and ceilingz > z then
				ceilingz = z
			end
			continue
		end

		local delta1 = z - (bottomheight + ((topheight - bottomheight)/2))
		local delta2 = thingtop - (bottomheight + ((topheight - bottomheight)/2))
		if bottomheight < ceilingz and abs(delta1) > abs(delta2) then
			ceilingz = bottomheight
		end
	end

	return ceilingz
end

local function L_AdjustMobjToSlope(moth)
	if not runscript then return end
	local sec = R_PointInSubsector(moth.x, moth.y).sector
	local zh = 0 // spawnpoint height
	if moth.spawnpoint then
		zh = moth.spawnpoint.options << 12
	end
	local ch = L_CeilingzAtPos(moth.x, moth.y, moth.z, moth.height)
	local fh = P_FloorzAtPos(moth.x, moth.y, moth.z, moth.height)
	
	if moth.flags2 & MF2_OBJECTFLIP then
		if not sec.c_slope then return end
		moth.z = ch - moth.height - zh
	else
		if not sec.f_slope then return end
		moth.z = fh + zh
	end
end

local function L_AdjustMapthingToSlope(moth)
	// this adjust mapthings so objects that respawn, respawn in the correct position
	local sec = R_PointInSubsector(moth.x*FRACUNIT, moth.y*FRACUNIT).sector
	local zh = moth.options>>4 // thing z height
	local height = 16*FRACUNIT
	if moth.mobj then
		height = moth.mobj.height
	end
	local ch = L_CeilingzAtPos(moth.x*FRACUNIT, moth.y*FRACUNIT, moth.z*FRACUNIT, height)
	local fh = P_FloorzAtPos(moth.x*FRACUNIT, moth.y*FRACUNIT, moth.z*FRACUNIT, height)
	if moth.options & 2 then // flip
		if not sec.c_slope then return end
		moth.z = ch - zh - height
	else
		if not sec.f_slope then return end
		moth.z = fh + moth.z
	end
end

addHook("MapLoad", do runscript = mapheaderinfo[gamemap].copyslopefix ~= nil end)
addHook("NetVars", function(net) runscript = net($) end) // might as well be paranoid
addHook("ThinkFrame", do
	if runscript and leveltime == 2 then
		// MapThingSpawn where are you
		for thing in mapthings.iterate do
			L_AdjustMapthingToSlope(thing)
		end
		runscript = false // only run once
	end
end)
addHook("MobjSpawn", L_AdjustMobjToSlope)

rawset(_G, "copyslopefix_initialized", true)